import axios from 'axios';
import * as signalR from "@microsoft/signalr";
import adminPrerequisite from './adminPrerequisite.js';
import { NodeId, ReadNode, CreateNode } from './types.js';
import { IWebError, IWebResponse, WebResponseStatus } from './web-response.js';
import { NodeQueryIdentifier } from './node.js';
import { IUpdateNode } from './update-node.js';
import { INodeChildrenChanges, INodePropertyChanges, INodeValueChanges } from './receive-changes.js';
import { INodeValue, INodeValueStatus } from './node-value.js';

const signalrNodesPath = '/api/signalr/nodes';
const authenticatePath = '/api/rest/users/authenticate';

interface AuthenticateModel {
  username: string;
  password: string;
  adminPrerequisite?: string;
}

export class Client {
  public constructor(host: string) {
    this.host = host;
  }

  public async connect(usernameOrToken: string, password?: string) {
    if (password === undefined) {
      this.authToken = usernameOrToken;
    }
    else {
      this.username = usernameOrToken;
      this.password = password;
    }

    if (this.authToken === undefined) {
      let credentials: AuthenticateModel = {
        username: this.username,
        password: this.password
      };
      
      if (this.username === 'admin') {
        credentials.adminPrerequisite = adminPrerequisite;
      }

      let response = await axios.post(
        this.host + authenticatePath,
        credentials);

      if (response.status === 200) {
        this.authToken = response.data.token;
      }
      else {
        throw new Error('An error occurred during authentication: ' + response.data);
      }
    }

    this.connection = new signalR.HubConnectionBuilder()
      .withUrl(
        this.host + signalrNodesPath,
        { accessTokenFactory: () => this.authToken })
      .build();

    await this.connection.start();
  }

  public async disconnect() {
    await this.connection.stop();
  }

  public async readNode(id: NodeQueryIdentifier): Promise<ReadNode> {
    let request = {
      command: {
        id: id
      }
    };
    let response = await this.connection.invoke('ReadNode', request);
    this.handleError(response);
    return response.result.node;
  }

  public async readNodes(ids: Array<NodeQueryIdentifier>): Promise<ReadNode[]> {
    let request = {
      commands: ids.map(e => {
        return { id: e }
      })
    };
    let response = await this.connection.invoke('ReadNodes', request);
    this.handleError(response);
    return response.result.map(e => e.node);
  }

  public async createNode(node: CreateNode): Promise<NodeId> {
    let request = {
      command: {
        node: node
      }
    };
    let response = await this.connection.invoke('CreateNode', request);
    this.handleError(response);
    return response.result.id;
  }

  public async deleteNode(id: NodeQueryIdentifier): Promise<void> {
    let request = {
      command: {
        id: id
      }
    };
    let response = await this.connection.invoke('DeleteNode', request);
    this.handleError(response);
  }

  public async updateNode(id: NodeQueryIdentifier, node: IUpdateNode): Promise<void> {
    let request = {
      command: {
        id: id,
        node: node
      }
    };
    let response = await this.connection.invoke('UpdateNode', request);
    this.handleError(response);
  }

  public async linkNode(id: NodeQueryIdentifier, destinationId: NodeQueryIdentifier): Promise<void> {
    let request = {
      command: {
        id: id,
        destinationId: destinationId
      }
    };
    let response = await this.connection.invoke('LinkNode', request);
    this.handleError(response);
  }

  public async readChildNodes(id: NodeQueryIdentifier): Promise<ReadNode> {
    let request = {
      command: {
        id: id
      }
    };
    let response = await this.connection.invoke('ReadChildNodes', request);
    this.handleError(response);
    return response.result.nodes;
  }

  public subscribeValueChanges(
      ids: Array<NodeQueryIdentifier>,
      createSubscription?: boolean,
      interval?: number,
      forceUpdate?: boolean): signalR.IStreamResult<INodeValueChanges> {
    let request = {
      nodes: ids,
      createSubscription: createSubscription,
      interval: interval,
      forceUpdate: forceUpdate
    };
    return this.connection.stream(
      'SubscribeValueChanges',
      request
    )
  }

  public subscribePropertyChanges(...ids: Array<NodeQueryIdentifier>): signalR.IStreamResult<INodePropertyChanges> {
    let request = {
      nodes: ids
    };
    return this.connection.stream(
      'SubscribePropertyChanges',
      request
    )
  }

  public subscribeChildrenChanges(...ids: Array<NodeQueryIdentifier>): signalR.IStreamResult<INodeChildrenChanges> {
    let request = {
      nodes: ids
    };
    return this.connection.stream(
      'SubscribeChildrenChanges',
      request
    )
  }

  public async writeNodeValue(id: NodeQueryIdentifier, value: INodeValue | any): Promise<INodeValueStatus> {
    if (!value || !value.value) {
      value = {
        value: value
      }
    };
    let request = {
      command: {
        id: id,
        value: value
      }
    };
    let response = await this.connection.invoke('WriteNodeValue', request);
    this.handleError(response);
    return response.result.status;
  }

  public async readNodeValue(id: NodeQueryIdentifier, performRead: boolean = false, timeToLive?: number): Promise<INodeValue> {
    let request = {
      command: {
        id: id,
        performRead: performRead,
        timeToLive: timeToLive
      }
    };
    let response = await this.connection.invoke('ReadNodeValue', request);
    this.handleError(response);
    return response.result.value;
  }

  public async readNodeValueHistory(id: NodeQueryIdentifier, start: Date, end: Date, timeInterval?: number): Promise<Array<INodeValue>> {
    let request = {
      command: {
        id: id,
      },
      start: start,
      end: end,
      timeInterval: timeInterval
    };
    let response = await this.connection.invoke('ReadNodeValueHistory', request);
    this.handleError(response);
    return response.result.values;
  }

  public async readNodesValueHistory(ids: Array<NodeQueryIdentifier>, start: Date, end: Date, timeInterval: number): Promise<Array<Array<INodeValue>>> {
    let request = {
      command: ids.map(e => {
        return { id: e };
      }),
      start: start,
      end: end,
      timeInterval: timeInterval
    };
    let response = await this.connection.invoke('ReadNodesValueHistory', request);
    this.handleError(response);
    return response.result.map(e => e.values);
  }

  private handleError(response: IWebResponse) {
    if (response.status === WebResponseStatus.success) {
      return response;
    }
    else {
      let message = Client.generateErrorMessage(response.error);
      throw new Error(message);
    }
  }

  private static generateErrorMessage(error: IWebError) {
    let message = error.detail;
    if (error.parameters) {
      console.log(error.parameters);
      message += ':\n\nParameters:\n';
      for (const [key, value] of Object.entries(error.parameters)) {
      // error.parameters.forEach((value, key) => {
        message += `"${key}": ${JSON.stringify(value)}\n`;
      }
      message += '\n\n';
    }
    if (error.errors) {
      message += ':\n\nSub errors:\n';
      error.errors.forEach(e => {
        message += this.generateErrorMessage(e);
      })
    }
    return message;
  }

  private host: string;
  private username: string;
  private password: string;
  private authToken: string;
  private connection: signalR.HubConnection;
}

export * from './types.js';
