import * as codabix from './lib/codabix-signalr.js';

async function createNodes(client: codabix.Client): Promise<Array<codabix.NodeId>> {
  let nodes: Array<codabix.CreateNode> = [
    new codabix.CreateFolderNode({ name: 'TestNode', parentId: '/Nodes' }),
    new codabix.CreateValueNode({ valueType: codabix.Type.int16, name: 'A node', parentId: '/Nodes/TestNode' }),
    new codabix.CreateValueNode({ valueType: codabix.Type.int16, name: 'Another node', parentId: '/Nodes/TestNode' }),
    new codabix.CreateFolderNode({ name: 'A folder node', parentId: '/Nodes/TestNode' }),
    new codabix.CreateMethodNode({ name: 'A method node', parentId: '/Nodes/TestNode' }),
    new codabix.CreateFolderNode({ name: 'Link', parentId: '/Nodes/TestNode' }),
    new codabix.CreateFolderNode({ name: 'LinkTo', parentId: '/Nodes/TestNode' }),
  ];
  let nodeIds: Array<codabix.NodeId> = [];
  for (let i = 0; i < nodes.length; i++) {
    try {
      let id = await client.createNode(nodes[i]);
      nodeIds.push(id);
    }
    catch(e) {
      console.log(e);
    }
  }
  return nodeIds;
}

async function readNodes(client: codabix.Client, nodeIds: Array<codabix.NodeId>) {
  let nodes = await client.readNodes(nodeIds);
  console.log(nodes);
}

async function updateNode(client: codabix.Client) {
  let updateNode: codabix.IUpdateNode = {
    path: 'this is the node path',
  };
  let nodePath = '/Nodes/TestNode/A node';
  await client.updateNode(nodePath, updateNode);
  let node = await client.readNode(nodePath);
  console.log('Updated node path: ' + node.path);
}

async function readNodeValue(client: codabix.Client) {
  let nodePath = '/Nodes/TestNode/A node';
  let value = await client.readNodeValue(nodePath);
  console.log('Node value: ' + JSON.stringify(value));
}

async function writeNodeValue(client: codabix.Client) {
  let nodePath = '/Nodes/TestNode/A node';
  let status = await client.writeNodeValue(nodePath, true);
  console.log('status: ' + JSON.stringify(status));
}

async function readChildNodes(client: codabix.Client) {
  let nodePath = '/Nodes/TestNode';
  let nodes = await client.readChildNodes(nodePath);
  console.log(`Child nodes of ${nodePath}: `)
  console.log(nodes);
}

async function linkNode(client: codabix.Client) {
  let nodePath = '/Nodes/TestNode/Link';
  let destinationPath = '/Nodes/TestNode/LinkTo';
  await client.linkNode(nodePath, destinationPath);
}

async function main() {
  let client = new codabix.Client('http://localhost:8181');
  try {
    await client.connect('demo@user.org', 'demo');
    let nodeIds = await createNodes(client);
    
    await readNodes(client, nodeIds);
    await updateNode(client);
    await readNodes(client, nodeIds);
    await readChildNodes(client);
    await readNodeValue(client);
    await writeNodeValue(client);
    await readNodeValue(client);
    await linkNode(client);
    await readNodes(client, nodeIds);

    let subscription = client.subscribeChildrenChanges('/Nodes/TestNode/Link')
      .subscribe({
        next: (e) => {
          console.log("\Children change event:\n");
          console.log("Node: " + e.node);
          console.log("added: " + JSON.stringify(e.added));
          console.log("removed: " + JSON.stringify(e.removed));
        },
        complete: () => {
          console.log("'ReceiveChildrenChanges' has been completed.");
        },
        error: (err) => {
          console.log("An error occurred on 'ReceiveChildrenChanges':\n" + err);
        }
    });
    
    await client.createNode(new codabix.CreateFolderNode({ name: 'link childnode', parentId: '/Nodes/TestNode/LinkTo' }));

    let subscription2 = client.subscribeChildrenChanges('/Nodes/TestNode/Link/link childnode')
      .subscribe({
        next: (e) => {
          console.log("\Children change event:\n");
          console.log("Node: " + e.node);
          console.log("added: " + JSON.stringify(e.added));
          console.log("removed: " + JSON.stringify(e.removed));
        },
        complete: () => {
          console.log("'ReceiveChildrenChanges' has been completed.");
        },
        error: (err) => {
          console.log("An error occurred on 'ReceiveChildrenChanges':\n" + err);
        }
    });

    let id2 = await client.createNode(new codabix.CreateValueNode({ name: 'A second value node', valueType: codabix.Type.int16, parentId: '/Nodes/TestNode/Link/link childnode' }));
    let node = await client.readNode(id2);
    console.log(node);
    await client.deleteNode(id2);


    setTimeout(() => {
      subscription.dispose();
      subscription2.dispose();
    }, 120000);
  }
  catch(e) {
    console.log(e);
  }
  finally {
    await client.deleteNode('/Nodes/TestNode');
  }
}

main();
